from cmath import sqrt
from typing import Tuple, overload


class Vector2():
    """
    2D Vector.
    """

    x : float = 0.0
    """
    The x component of the vector.
    """

    y : float = 0.0
    """
    The y component of the vector.
    """


    __magnitude : float = 0.0
    """
    The length of the vector.
    """

    __calculatedMagnitude : float = False
    """
    Whether the magnitude is calculated. 
    """


    __sqrMagnitude : float = 0.0
    """
    The square of the length of the vector.
    """


    __calculatedSqrMagnitude : bool = False
    """
    Whether the sqrMagnitude is calculated. 
    """


    def __init__(self, x : float, y : float):
        self.x = x
        self.y = y
        pass


    def __add__(self, other):
        return Vector2(self.x + other.x, self.y + other.y)


    def __sub__(self, other):
        return Vector2(self.x - other.x, self.y - other.y)

    
    def __str__(self) -> str:
        return "(%f, %f)" % (self.x, self.y)


    @staticmethod
    def zero():
        return Vector2(0, 0)


    def magnitude(self):
        if not self.__calculatedMagnitude:
            self.__sqrMagnitude = self.x * self.x + self.y * self.y
            self.__calculatedMagnitude = True
        return self.__magnitude


    def normalized(self):
        if self.magnitude == 0:
            return Vector2.zero()
        return Vector2(self.x / self.magnitude, self.y / self.magnitude)


    def sqrMagnitude(self):
        if not self.__calculatedSqrMagnitude:
            self.__magnitude = sqrt(self.__sqrMagnitude)
            self.__calculatedSqrMagnitude = True
        return self.__sqrMagnitude


    def getTuple(self):
        return (self.x, self.y)
    